from importlib import import_module
from typing import Union

from s3_parse_url import parse_s3_url
from s3_parse_url.base import S3DataSource


def import_module_if_installed(package_name: str):
    try:
        return import_module(package_name)
    except ImportError:
        return None


Parseable = Union[str, S3DataSource]
aiobotocore = import_module_if_installed("aiobotocore")
boto3 = import_module_if_installed("boto3")


def get_boto_client_kwargs(url: Parseable, **kwargs):
    if isinstance(url, S3DataSource):
        datasource = url
    else:
        datasource = parse_s3_url(url)
    return {
        "aws_access_key_id": datasource.access_key_id,
        "aws_secret_access_key": datasource.secret_access_key,
        "endpoint_url": datasource.endpoint_url,
        "region_name": datasource.region,
    }


def get_aiobotocore_client(url: str, **kwargs):
    if aiobotocore is None:
        raise ValueError("aiobotocore not installed")

    from aiobotocore.session import get_session

    return get_session().create_client(
        "s3", **get_boto_client_kwargs(url, **kwargs),
    )


def get_boto3_client(url: str, **kwargs):
    if boto3 is None:
        raise ValueError("boto3 not installed")
    return boto3.client("s3", **get_boto_client_kwargs(url, **kwargs))



from s3_parse_url.exceptions import StorageAlreadyRegistered
from s3_parse_url.storages.amazon import AmazonS3
from s3_parse_url.storages.mail_ru import HotBox, IceBox
from s3_parse_url.storages.minio import Minio
from s3_parse_url.storages.selectel import SelectelStorage
from s3_parse_url.storages.yandex import YandexCloud

SUPPORTED_STORAGES = {}


def register_storage(schema: str, cls):
    if schema in SUPPORTED_STORAGES:
        raise StorageAlreadyRegistered("Schema registered")
    SUPPORTED_STORAGES[schema] = cls


register_storage("s3", AmazonS3)
register_storage("selectel", SelectelStorage)
register_storage("mailru", HotBox)
register_storage("mailru+hot", HotBox)
register_storage("mailru+ice", IceBox)
register_storage("minio", Minio)
register_storage("yandex", YandexCloud)



from s3_parse_url.base import S3DataSource


class YandexCloud(S3DataSource):
    default_endpoint = "https://storage.yandexcloud.net"
    default_region = "ru-central1"
    allowed_schemas = ["yandex"]



from s3_parse_url.base import S3DataSource


class IceBox(S3DataSource):
    default_region = "ru-msk"
    default_endpoint = "https://ib.bizmrg.com"
    allowed_schemas = ["mailru+ice"]


class HotBox(S3DataSource):
    default_region = "ru-msk"
    default_endpoint = "https://hb.bizmrg.com"
    allowed_schemas = ["mailru", "mailru+hot"]



from s3_parse_url.base import S3DataSource

# See https://docs.min.io/docs/aws-cli-with-minio.html
#
# A notable thing: according to the docs, the default region
# should be "us-east-1"


class Minio(S3DataSource):
    allowed_schemas = ["minio"]
    default_region = "us-east-1"
    allow_insecure_scheme = True

    def _parse_endpoint_url(self) -> str:
        endpoint_url = super()._parse_endpoint_url()
        return endpoint_url



from urllib.parse import urlparse

from s3_parse_url.base import INSECURE_SCHEME_SUFFIX, S3DataSource
from s3_parse_url.exceptions import UnsupportedStorage
from s3_parse_url.storages import SUPPORTED_STORAGES

__version__ = "0.2.1"
__all__ = ["parse_s3_dsn", "parse_s3_url", "UnsupportedStorage"]


def parse_s3_dsn(dsn: str) -> S3DataSource:
    """
    Parses a datasource string to a dict of arguments compatible with
    """
    insecure_suffix = INSECURE_SCHEME_SUFFIX
    try:
        scheme = urlparse(dsn).scheme.lower()
        if scheme.endswith(insecure_suffix):
            scheme = scheme[:-len(insecure_suffix)]
        cls = SUPPORTED_STORAGES[scheme]
    except (AttributeError,
            KeyError):
        raise UnsupportedStorage()
    else:
        return cls(dsn)


parse_s3_url = parse_s3_dsn



from s3_parse_url.base import S3DataSource


class AmazonS3(S3DataSource):
    allowed_schemas = ["s3"]
    default_region = "us-east-1"
    amazon_s3_domains = (
        ".s3.amazonaws.com",
        ".s3.aws.amazon.com",
    )

    def _is_amazon_host(self):
        return all((
            self._raw_bits.hostname,
            self._raw_bits.hostname.lower().endswith(self.amazon_s3_domains)
        ))

    def _strip_amazon_prefix(self):
        for d in self.amazon_s3_domains:
            if not self._raw_bits.hostname.endswith(d):
                continue
            return self._raw_bits.hostname[:-len(d)]

    def _is_host_given(self):
        if self._is_amazon_host():
            return False
        return super()._is_host_given()

    def _parse_bucket_name(self) -> str:
        _parent = super()._parse_bucket_name()
        if self._is_amazon_host():
            return self._strip_amazon_prefix()
        return _parent



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['s3_parse_url', 's3_parse_url.ext', 's3_parse_url.storages']

package_data = \
{'': ['*']}

setup_kwargs = {
    'name': 's3-parse-url',
    'version': '0.2.1',
    'description': 'A small tool to parse URLs for S3 compatible storage services',
    'long_description': '############\ns3-parse-url\n############\n\n.. image:: https://img.shields.io/pypi/pyversions/s3-parse-url\n  :alt: PyPI - Python Version\n\n.. image:: https://img.shields.io/pypi/v/s3-parse-url\n  :alt: PyPI\n\n.. image:: https://img.shields.io/pypi/l/s3-parse-url\n :alt: PyPI - License\n\n\n.. image:: https://coveralls.io/repos/github/marazmiki/s3-parse-url/badge.svg?branch=master\n :target: https://coveralls.io/github/marazmiki/s3-parse-url?branch=master\n\n.. image:: https://img.shields.io/codacy/grade/80c1a1af099848ddb5cc86221723f8d5\n  :alt: Codacy grade\n\n-----\n\nParses S3 credentials from the given string and returns it in comfortable\nformat to pass to popular clients like boto3.\n\nAbout\n=====\n\nThis is a small utility to parse locations of buckets of S3-compatible\nstorage (and, of course, the original Amazon\'s S3 itself) given in the URL form\nlike ``s3://bucket/``.\n\nIt could be useful in our epoch of the 12-factor applications when it\'s a\ngood practice to store credentials inside of environment variables.\n\nAlso, these days, there are some notable S3-compatible storage services:\n\n* `Selectel <https://>`_\n\n* `MinIO <https://min.io>`_ `(a self-hosted solution extremely handy for testing)`\n\nAnd dozens of others.\n\nWith ``s3-parse-url``, you can use any one of these services with no doubts about\nconfiguration endpoints. For example, you can connect to your Selectel storage\nwith ``boto3`` just using ``selectel://my-bucket`` DSN.\n\nThat\'s an example, what it was all about:\n\n.. code:: python\n\n    from s3_parse_url import s3_parse_url\n    from s3_parse_url.ext.clients import get_boto3_client\n\n    dsn = "s3://AKIA***OO:XP***@my-bucket/?region=us-east-2"\n\n    # It\'s a completely ready boto3 client instance to work with Selectel\n    s3_client = get_boto3_client(dsn)\n\nOf course, in the code above we worked with Selectel (have you ever heard\nabout it?). You can work this way with any S3 compatible storage. If you\nprefer unknown storage, you can easily create a plugin to add support for\nyour favorite service. Or, if you are a pervert, you can use a universal ``S3://``\nscheme, but in this case, you should manage endpoints by yourself:\n\n.. code:: python\n\n    from s3_parse_url.ext.clients import get_boto3_client\n\n    # Also should work\n    dsn = "s3://my-minio-user:my-minio-pass@minio.example.com:9000/?region=us-east-1"\n\n    # A ready client to work with a minio instance\n    s3_client = get_boto3_client(dsn)\n\n\nSupported providers\n===================\n\nCurrently we have support for these storages\n\n* Amazon S3\n* Selectel\n* Yandex\n* Mail.ru\n* MinIO\n\nBut you can easily add your own one.\n\nLicense\n=======\n\nThis project is licensed under the terms of the MIT license.\n\n',
    'author': 'Mikhail Porokhovnichenko',
    'author_email': 'marazmiki@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/marazmiki/s3-parse-url',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



class S3ParseException(Exception):
    pass


class UnsupportedStorage(S3ParseException):
    pass


class StorageAlreadyRegistered(S3ParseException):
    pass


class InsecureNotAllowed(S3ParseException):
    pass



from typing import List, Optional
from urllib.parse import parse_qs, urlparse

from s3_parse_url.exceptions import InsecureNotAllowed, UnsupportedStorage

SECURE_HTTP_SCHEME = "https"
INSECURE_HTTP_SCHEME = "http"
INSECURE_SCHEME_SUFFIX = "+insecure"


class S3DataSource:
    allowed_schemas: List[str] = ["s3"]
    default_endpoint: str = None
    default_region: str = None
    allow_insecure_scheme: bool = False
    region_param_name: str = "region"

    def __init__(self, dsn: str):
        self._dsn = dsn
        self._raw_bits = None
        self._parsed_bits = {}
        self._parse()

    def __str__(self):
        return self._dsn

    @property
    def endpoint_url(self) -> Optional[str]:
        return self._parsed_bits["endpoint_url"]

    @property
    def bucket_name(self) -> str:
        return self._parsed_bits["bucket_name"]

    @property
    def region(self) -> str:
        return self._parsed_bits["region"]

    @property
    def access_key_id(self) -> str:
        return self._parsed_bits["aws_access_key_id"]

    @property
    def secret_access_key(self) -> str:
        return self._parsed_bits["aws_secret_access_key"]

    @property
    def key(self) -> str:
        return self._parsed_bits["key"]

    def _parse(self):
        self._raw_bits = urlparse(self._dsn, allow_fragments=False)
        self._check_if_compatible()
        self._parsed_bits.update(**{
            "aws_secret_access_key": self._parse_secret_access_key(),
            "aws_access_key_id": self._parse_access_key_id(),
            "endpoint_url": self._parse_endpoint_url(),
            "bucket_name": self._parse_bucket_name(),
            "region": self._parse_region_name(),
            "key": self._parse_key(),
        })

    def _parse_access_key_id(self):
        return self._raw_bits.username

    def _parse_secret_access_key(self):
        return self._raw_bits.password

    def _is_host_given(self):
        "Checks if in the DSN there is an explicitly given domain name"
        return any((
            self._raw_bits.port is not None,
            self._raw_bits.hostname.count(".") > 0
        ))

    def _parse_endpoint_url(self):
        if not self._is_host_given():
            return self.default_endpoint
        schema_prefix = self._parse_schema_prefix()
        host, port = self._raw_bits.hostname, self._raw_bits.port
        if port:
            return schema_prefix + host + ":" + str(port) + "/"
        else:
            return schema_prefix + host + "/"

    def _parse_region_name(self) -> str:
        region = parse_qs(self._raw_bits.query).get(self.region_param_name)
        if region and region[0]:
            return region[0]
        else:
            return self.default_region

    def _parse_bucket_name(self) -> str:
        if self._is_host_given():
            bucket_name, *key_bits = self._raw_bits.path.lstrip("/").split("/")
            return bucket_name
        else:
            return self._raw_bits.hostname

    def _parse_key(self) -> str:
        key = self._raw_bits.path.lstrip("/")
        return key

    def _parse_schema_prefix(self) -> str:
        if self.allow_insecure_scheme:
            schema = (INSECURE_HTTP_SCHEME
                      if self._raw_bits.scheme.endswith(INSECURE_SCHEME_SUFFIX)
                      else SECURE_HTTP_SCHEME)
        else:
            schema = SECURE_HTTP_SCHEME
        return schema + "://"

    def _check_if_compatible(self):
        scheme = self._raw_bits.scheme
        if not self.allowed_schemas:
            return
        if not scheme:
            raise UnsupportedStorage("Unable to detect schema")

        canonical_scheme = scheme.lower()
        scheme_is_insecure = scheme.endswith(INSECURE_SCHEME_SUFFIX)

        if scheme_is_insecure:
            canonical_scheme = canonical_scheme[:-len(INSECURE_SCHEME_SUFFIX)]
            if not self.allow_insecure_scheme:
                raise InsecureNotAllowed("Insecure interaction does not "
                                         "allowed for the backend")
        if canonical_scheme not in self.allowed_schemas:
            raise UnsupportedStorage("Unsupported schema " + scheme +
                                     " for the backend")
